#ifndef PLS_TASK_H
#define PLS_TASK_H

#include "pls/internal/scheduling/continuation.h"
#include "pls/internal/scheduling/cont_manager.h"

namespace pls {
namespace internal {
namespace scheduling {

/**
 * A task to be executed by the runtime system.
 * Tasks are guaranteed to be executed exactly once.
 *
 * Override the execute_internal() method for your custom code.
 */
class base_task {
 protected:
  base_task() = default;

  /**
   * Overwrite this with the actual behaviour of concrete tasks.
   */
  virtual void execute_internal() = 0;

 public:
  void execute() {
    // TODO: Figure out slow path execution
    execute_internal();
  }
};

template<typename F, typename R1, typename R2, typename CF>
class task : public base_task {
  using continutation_type = continuation<R1, R2, CF>;

 public:
  template<typename FARG>
  explicit task(FARG &&function, continuation_node *continuation_node)
      : base_task{}, function_{std::forward<FARG>(function)}, continuation_node_{continuation_node} {}

  void execute_internal() override {
    auto *continuation = dynamic_cast<continutation_type *>(continuation_node_->continuation());
    continuation->store_result_2(function_());
    // TODO: Properly notify continuation on slow path
  }

 private:
  F function_;
  continuation_node *continuation_node_;
};

}
}
}

#endif //PLS_TASK_H
